;; $Id: ch5-7-1.scm,v 1.19 1999/05/05 03:22:19 leavens Exp $

;; Code for EOPL section 5.7.1 (dynamic scoping)
;; using the chararacter string parser.

;;; Grammar:
;;	    <exp> ::= <integer-literal>
;;		    | <varref>
;;		    | <operator> <operands>
;;                  | if <exp> then <exp> else <exp>
;;	            | let <decls> in <exp>
;;	            | proc <varlist> <exp>
;;	            | <var> := <exp>
;;	<operator> ::= <varref> | (<exp>)
;;	<operands> ::= ( )
;;	            | (<operand> {, <operand>}*)
;;	 <operand> ::= <exp>
;;	  <varref> ::= <var>
;;	<decls> ::= <decl> {;<decl>}*
;;	 <decl> ::= <var> = <exp>
;;	<varlist> ::= ( ) | ( <vars> )
;;	 <vars> ::= <var> {,<var>}*

;;; Syntactic sugars
;;       none

(load-quietly-from-lib "syntax-expand-0.scm")

;;; Top-level Forms
;;      <form> ::= <exp>

(load-quietly-from-lib "eval-form-0.scm")

;;; Parsing
;; To use "run", you also need to do
;;        (load "run.scm")
;; To use "read-eval-print", you also need to do
;;        (load "rep.scm")
;; Note that you can't use both of these at the same time.

(load-quietly-from-lib "undefine-read-eval-print.scm")
(load-quietly-from-lib "undefine-run.scm")
(load-quietly-from-lib "ch5-char-parser.scm")

;;; Domains for this interpreter:
;;;
;;; Environment = <var> -> Denoted-Value
;;; Denoted-Value = Cell(Expressed-Value)
;;; Expressed-Value = Number + Procedure-Text + List(Expressed-Value) + Void
;;; Procedure-Text = prim-proc + proc

;;; ADTs used

(load-quietly-from-lib "environments.scm")
(load-quietly-from-lib "procedure-text.scm")

;;; Helpful stuff

(load-quietly-from-lib "compose.scm")
(load-quietly-from-lib "ignore.scm")

;;; Truth values

;; Anonymous figure : page 146

(define true-value?
  ;; TYPE: (-> (number) boolean)
  (lambda (x)
    (not (zero? x))))

(define boolean->int
  ;; TYPE: (-> (boolean) integer)
  (lambda (x)
    (if  x  1  0)))

;;; The following have to be after all the specific domains are defined.
(load-quietly-from-lib "ch5-7-1-denoted-value.scm")
(load-quietly-from-lib "ch5-7-1-expressed-value.scm")

;;; Figure 5.7.1 : page 169

(define eval-exp
  ;; TYPE: (-> (parsed-exp Environment) Expressed-Value)
  (lambda (exp env)
    (variant-case exp
      (lit (datum) (number->expressed datum))
      (varref (var) (denoted->expressed (apply-env env var)))
      (app (rator rands)
           (let ((proc (expressed->procedure-text (eval-exp rator env)))
		 (args (eval-rands rands env)))
	     (apply-proc proc args env)))
      (if (test-exp then-exp else-exp)
	  (if (true-value? (expressed->number (eval-exp test-exp env)))
	      (eval-exp then-exp env)
	      (eval-exp else-exp env)))
      (proc (formals body)
	    (procedure-text->expressed (proc->procedure-text exp)))
      (varassign (var exp)
		 (void->expressed (cell-set! (apply-env env var)
					     (eval-exp exp env))))
      (begin (exp1 exp2)
	(begin
	  (ignore (eval-exp exp1 env))
	  (eval-exp exp2 env)))
      ;; let syntax added and modified from Figure 5.4.1; solves problem 5.5.1
      (let (decls body)
        (let ((vars (map decl->var decls))
              (exps (map decl->exp decls)))
          (let ((new-env (extend-env vars
				     (map (lambda (r)
					    (expressed->denoted
					     (eval-exp r env)))
					  exps)
				     env)))
            (eval-exp body new-env))))
      (else (error "Invalid abstract syntax: " exp)))))

(define eval-rands
  ;; TYPE: (-> ((list parsed-exp) Environment) (list Expressed-Value))
  (lambda (rands env)
    (map (lambda (rand) (eval-exp rand env))
         rands)))

;;; Figure 5.7.1 : page 169

(define apply-proc
  ;; TYPE: (-> (Procedure-Text (list Expressed-Value) Environment)
  ;;           Expressed-Value)
  (lambda (proc args current-env)
    (variant-case proc
      (prim-proc (prim-op) (apply-prim-op prim-op args))
      (proc (formals body)
        (eval-exp body 
          (extend-env formals (map expressed->denoted args) current-env)))
      (else (error "Invalid procedure-text:" proc)))))

(define apply-prim-op
  ;; TYPE: (-> (symbol (list Expressed-Value)) Expressed-Value)
  (lambda (prim-op args)
    (case prim-op
      ((+) (number->expressed (+ (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((*) (number->expressed (* (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((-) (number->expressed (- (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((add1) (number->expressed (+ (expressed->number (car args)) 1)) )
      ((sub1) (number->expressed (- (expressed->number (car args)) 1)) )
      ;; primitive operators added to interpreter follow
      ((minus) (number->expressed (- (expressed->number (car args)))))
      ((cons) (list->expressed (cons (car args)
				     (expressed->list (cadr args)))))
      ((car) (car (expressed->list (car args))))
      ((cdr) (list->expressed (cdr (expressed->list (car args)))))
      ((cadr) (cadr (expressed->list (car args))))
      ((caddr) (caddr (expressed->list (car args))))
      ((cadddr) (cadddr (expressed->list (car args))))
      ((list) (list->expressed args))
      ((null) (number->expressed (boolean->int
				  (null? (expressed->list (car args))))) )
      ((eq) (number->expressed (boolean->int (eq? (car args) (cadr args)))) )
      ((equal) (number->expressed (boolean->int
				   (= (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((less) (number->expressed (boolean->int
				   (< (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((greater) (number->expressed (boolean->int
				   (> (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((zero) (number->expressed (boolean->int
				  (zero? (expressed->number (car args))))) )
      ((write) (void->expressed (write (car args))))
      ((print) (void->expressed (apply displayln args)))
      ((writeln) (void->expressed (begin (for-each (lambda (arg)
						     (write arg)
						     (display " "))
						   args)
					 (newline))))
      ((space) (void->expressed (display " ")))
      ((newline) (void->expressed (newline)))
      ((printchar) (void->expressed
		    (display (integer->char
			      (expressed->number (car args))))))
      (else (error "Invalid prim-op name:" prim-op)))))

(define prim-op-names  ; TYPE: (list symbol)
  ;; modified to add primitive operators
  '(+ - * add1 sub1 minus cons car cdr list null eq equal less greater zero
      cadr caddr cadddr write print writeln space newline
      printchar))

(define init-env       ; TYPE: Environment
  (extend-env
    prim-op-names
    (map (compose expressed->denoted
		  (compose procedure-text->expressed make-prim-proc))
	 prim-op-names)
    (extend-env
      '(emptylist)
      (list (expressed->denoted (list->expressed '())))
      the-empty-env)))

;;; something to make the read-eval-print loop work...

(define closure?
  ; TYPE: (-> (datum) boolean)
  (lambda (x) #f))
