;;; $Id: char-parser-extended.scm,v 1.7 1999/04/26 20:15:44 leavens Exp $
;;;
;;; This is an extended parser that adds syntax for various exercises
;;; See the keywords below, and the parser productions at the end of the file.
;;; 

(load-quietly-from-lib "char-parser.scm")
(load-quietly-from-lib "parsed-exp-extended.scm")

;;; new keywords

(set! keywords-list
  (append
   keywords-list
   ;; additional keywords for various exercises
   '(unless when cond case of repeat for foreach to do while until
     run multiassign from and or implies within use with)
   ))


;;; get-constructor-from-name revised from Figure D.3 : page 476, 477

(define get-constructor-from-name
  (lambda (prod-name)
    (case prod-name
      ((chain) (lambda (x) x))
      ((lit) make-lit)      
      ((app) make-app)
      ((app/var) (lambda (var rands) (make-app (make-varref var) rands)))
      ((begin) (letrec ((loop (lambda (exps)
                                (if (null? (cdr exps))
                                    (car exps)
                                    (make-begin (car exps) (loop (cdr exps)))))))
                 loop))
      ((define) make-define)
      ((varref) make-varref)
      ((if) make-if)
      ((proc) make-proc)
      ((let) make-let)
      ((letmutable) make-letmutable)
      ((letrecproc) make-letrecproc)
      ((letrec) make-letrec)
      ((decl) make-decl)
      ((procdecl) make-procdecl)
      ((varassign) make-varassign)
      ((dynassign) make-dynassign)
      ((letdynamic) make-letdynamic)
      ((definearray) make-definearray)      
      ((arrayref) make-arrayref)      
      ((arrayref/var) (lambda (var index)
                        (make-arrayref (make-varref var) index)))
      ((arrayassign) make-arrayassign)
      ((arrayassign/var) (lambda (var index val)
                           (make-arrayassign (make-varref var) index val)))
      ((letarray) make-letarray)
      ((letproc) make-letproc)
      ((local) make-local)
      ((stuff-and-keydecls) append)      
      ((keydecl) make-keydecl)
      ((i-var) make-i-varref)
      ((c-var) make-c-varref)
      ((method) make-method)
      ((letcont) make-letcont)
      ((i-varassign) make-i-varassign)
      ((c-varassign) make-c-varassign)
      ((new-simpleinst) make-new-simpleinst)
      ((new-simpleclass) make-new-simpleclass)
      ((new-class) make-new-class)
      ((instance) make-new-instance)
      ((meth-app) (lambda (var rands)
                    (if (null? rands)
                        (error "Invalid method rands: " rands)
                        (let ((rand (car rands)))
                          (if (and (varref? rand)
                                   (eq? (varref->var rand) 'super))
                              (make-super-meth-app var (cdr rands))
                              (make-meth-app var rands))))))
      ((abort) make-abort)
      ((callcc) make-callcc)
      ((coroutine) make-coroutine)
      ((wind) make-wind)
      ((sum) make-sum)
      ;; new stuff follows
      ((unless) make-unless-exp)
      ((when) make-when-exp)
      ((cond) make-cond-exp)
      ((clause) make-clause)
      ((case) make-case-exp)
      ((one-case) make-one-case)
      ((repeat) make-repeat-exp)
      ((for) make-for)
      ((foreach) make-foreach)
      ((while) make-while)
      ((do-until) make-do-until)
      ((run-while) make-run-while)
      ((multiassign) make-multiassign)
      ((and) make-and-exp)
      ((or) make-or-exp)
      ((implies) make-implies-exp)
      ((within-use) make-within-use)
      ((with-do) make-with-do)
      (else (error "Bad production name: " prod-name)))))

;;; parse-exp revised from Figure D.2 : page 474, 475

(define parse-exp
  (lambda (token)
    ((case (token->class token)
      ((number) seen-number)
      ((variable) seen-variable)
      ((lparen) seen-lparen)
      ((if) seen-if)
      ((let) seen-let)
      ((proc) seen-proc)
      ((begin) seen-begin)
      ((letmutable) seen-letmutable)
      ((letrecproc) seen-letrecproc)
      ((letrec) seen-letrec)
      ((letdynamic) seen-letdynamic)
      ((letarray) seen-letarray)
      ((letproc) seen-letproc)
      ((local) seen-local)
      ((dollar-sign) seen-dollar-sign)
      ((ampersand) seen-ampersand)
      ((double-ampersand) seen-double-ampersand)
      ((method) seen-method)
      ((simpleclass) seen-simpleclass)      
      ((simpleinstance) seen-simpleinstance)
      ((class) seen-class) 
      ((instance) seen-instance)
      ((abort) seen-abort)
      ((letcont) seen-letcont)
      ((callcc) seen-callcc)
      ((coroutine) seen-coroutine)
      ((wind) seen-wind)
      ((sum) seen-sum)
      ;; new stuff follows below
      ((unless) seen-unless)
      ((when) seen-when)
      ((cond) seen-cond)
      ((case) seen-case)
      ((repeat) seen-repeat)
      ((for) seen-for)
      ((foreach) seen-foreach)
      ((while) seen-while)
      ((do) seen-do)
      ((run) seen-run)
      ((multiassign) seen-multiassign)
      ((and) seen-and)
      ((or) seen-or)
      ((implies) seen-implies)
      ((within) seen-within)
      ((with) seen-with)
      (else (error "Invalid parse-exp token: " token)))
     token)))

;;; New concrete syntax:
;;;
;;       <exp> ::= ...
;;                  | <unless>
;;                  | <when>
;;                  | <cond-exp>
;;                  | <case-exp>
;;                  | <repeat-exp>
;;                  | <for>
;;                  | <foreach>
;;                  | <while>
;;                  | <do-until>
;;                  | <run-while>
;;                  | <multiassign>
;;                  | <and-exp>
;;                  | <or-exp>
;;                  | <implies-exp>
;;                  | <within-use>
;;                  | <with-do>

;;; <unless> ::= unless <exp> do <exp>
(define seen-unless
  (lambda (token) 
    (check/drop 'unless
       (process-nt parse-exp
	  (check/drop 'do
	     (process-nt parse-exp
	        (reduce 'unless)))))))

;;; <when> ::= when <exp> do <exp>
(define seen-when
  (lambda (token) 
    (check/drop 'when
       (process-nt parse-exp
	  (check/drop 'do
	     (process-nt parse-exp
	        (reduce 'when)))))))

;;; <cond-exp> ::= cond <clauses> end
(define seen-cond
  (lambda (token) 
    (check/drop 'cond
       (process-nt parse-clauses
          (check/drop 'end
	    (reduce 'cond))))))

;;; <clauses> ::= <clause> | <clause> {; <clause>}*
(define parse-clauses
  (parse-loop* 'end 'semicolon
     (lambda (action) (process-nt parse-clause action))))

;;; <clause> ::= <exp> then <exp>
(define parse-clause
  (lambda (token)
    (process-nt parse-exp
       (check/drop 'then
	 (process-nt parse-exp
	    (reduce 'clause))))))

;;; <case-exp> ::= case <exp> of <one-cases> end
(define seen-case
  (lambda (token) 
    (check/drop 'case
      (process-nt parse-exp
	(check/drop 'of
          (process-nt parse-one-cases
            (check/drop 'end
	      (reduce 'case))))))))

;;; <one-cases> ::= <one-case> | <one-case> {; <one-case>}*
(define parse-one-cases
  (parse-loop* 'end 'semicolon
     (lambda (action) (process-nt parse-one-case action))))

;;; <one-case> ::= <integer-literal> : <exp>
(define parse-one-case
  (lambda (token)
     (check/shift 'number
	(check/drop 'colon
	   (process-nt parse-exp
	      (reduce 'one-case))))))

;;; <repeat-exp> ::= repeat <variable> := 1 to <exp> do <exp>
(define seen-repeat
  (lambda (token)
    (check/drop 'repeat
      (check/shift 'variable
        (check/drop 'assign-sym
	  (check/drop 'number
	    (check/drop 'to
	      (process-nt parse-exp
	        (check/drop 'do
	          (process-nt parse-exp
		    (reduce 'repeat)))))))))))

;;; <for> ::= for <variable> = <exp> to <exp> do <exp>
(define seen-for
  (lambda (token)
    (check/drop 'for
      (check/shift 'variable
        (check/drop 'eqsign
	  (process-nt parse-exp
	    (check/drop 'to
	      (process-nt parse-exp
	        (check/drop 'do
	          (process-nt parse-exp
		    (reduce 'for)))))))))))

;;; <foreach> ::= foreach <variable> in <exp> do <exp>
(define seen-foreach
  (lambda (token)
    (check/drop 'foreach
      (check/shift 'variable
        (check/drop 'in
	  (process-nt parse-exp
	     (check/drop 'do
	        (process-nt parse-exp
		   (reduce 'foreach)))))))))

;;; <while> ::= while <exp> do <exp>
(define seen-while
  (lambda (token)
    (check/drop 'while
      (process-nt parse-exp
	(check/drop 'do
	  (process-nt parse-exp
	    (reduce 'while)))))))

;;; <do-until> ::= do <exp> until <exp>
(define seen-do
  (lambda (token)
    (check/drop 'do
      (process-nt parse-exp
	(check/drop 'until
	  (process-nt parse-exp
	    (reduce 'do-until)))))))

;;; <run-while> ::= run <exp> while <exp>
(define seen-run
  (lambda (token)
    (check/drop 'run
      (process-nt parse-exp
	(check/drop 'while
	  (process-nt parse-exp
	    (reduce 'run-while)))))))

;;; <multiassign> ::= multiassign <varlist> := <operands>
(define seen-multiassign
  (lambda (token)
    (check/drop 'multiassign
      (check/drop 'lparen
        (process-nt parse-varlist
          (check/drop 'rparen
 	    (check/drop 'assign-sym
	      (check/drop 'lparen
	        (process-nt parse-rands
        	  (check/drop 'rparen
	            (reduce 'multiassign)))))))))))

;;; <and-exp> ::= and <exp> <exp>
(define seen-and
  (lambda (token)
    (check/drop 'and
      (process-nt parse-exp
	(process-nt parse-exp
	  (reduce 'and))))))

;;; <or-exp> ::= or <exp> <exp>
(define seen-or
  (lambda (token)
    (check/drop 'or
      (process-nt parse-exp
	(process-nt parse-exp
	  (reduce 'or))))))

;;; <implies-exp> ::= implies <exp> <exp>
(define seen-implies
  (lambda (token)
    (check/drop 'implies
      (process-nt parse-exp
	(process-nt parse-exp
	  (reduce 'implies))))))

;;; <within-use> ::= within <body> use <decls> end
;;; <body> ::= <exp>
;;; <decls> ::= <decl> {; <decl>}*
(define seen-within
  (lambda (token)
    (check/drop 'within
      (process-nt parse-exp
	(check/drop 'use
          (process-nt
	    (parse-loop+ 'end 'semicolon
	      (lambda (action) (process-nt parse-decl action)))
	    (check/drop 'end
	      (reduce 'within-use))))))))

;;; <with-do> ::= with <decls> do <body>
;;; <decls> ::= <decl> {; <decl>}*
;;; <body> ::= <exp>
(define seen-with
  (lambda (token)
    (check/drop 'with
      (process-nt
       	(parse-loop+ 'do 'semicolon
	  (lambda (action) (process-nt parse-decl action)))
	(check/drop 'do
          (process-nt parse-exp
	    (reduce 'with-do)))))))
