;; $Id: syntax-expand-letrec.scm,v 1.6 1999/08/11 16:28:14 leavens Exp $

;; syntax-expand that handles all the syntax,
;; even from the exercises, in chapters 5 and 6 of EOPL,
;; with the following...

;;; Syntactic sugars
;;        let x1 = e1; ...; xn = en in body
;;    ==>
;;        (proc (x1, ..., xn) body) (e1, ..., en)
;;
;;        letrec x1 = e1; ...; xn = en in body
;;    ==>
;;        let x1 = 0; ...; xn = 0
;;        in let g1 = e1; ...; gn = en
;;           in begin
;;                begin x1 := e1; begin ...; begin xn := en; 0 end ... end;
;;                body
;;              end
;;
;;        letrecproc p1(f1s) = e1; ...; pn(fns) = en in body
;;    ==>
;;        letrec p1 = proc(f1s) e1; ...; pn = proc(fns) en in body

;;; Grammar:
;;	    <exp> ::= <integer-literal>
;;		    | <varref>
;;		    | <operator> <operands>
;;	            | let <decls> in <exp>
;;                  | if <exp> then <exp> else <exp>
;;                  | begin exp1; exp2 end
;;                  | letmutable <decls> in <exp>
;;	            | proc <varlist> <exp>
;;	            | <var> := <exp>
;;                  | letrec <decls> in <exp>
;;                  | letrecproc <procdecls> in <exp>
;;                  | letdynamic <decls> in <exp>
;;                  | dynassign <var> := <exp> during <exp>
;;                  | letarray <arraydecls> in <exp>
;;		    | <array-exp> [<exp>]
;;	            | <array-exp> [<exp>] := <exp>
;;                  | letproc <procdecls> in <exp>
;;                  | local <decls> in <exp>
;;	<operator> ::= <varref> | (<exp>)
;;	<operands> ::= ( )
;;	            | (<operand> {, <operand>}*)
;;	 <operand> ::= <exp>
;;	  <varref> ::= <var>
;;	<decls> ::= <decl> {;<decl>}*
;;	 <decl> ::= <var> = <exp>
;;	<varlist> ::= ( ) | ( <vars> )
;;	 <vars> ::= <var> {,<var>}*
;;	<procdecls> ::= <procdecl> {;<procdecl>}*
;;	 <procdecl> ::= <var> <varlist> = <exp>
;;	   <array-exp> ::= <varref> | (<exp>)
;;	  <arraydecls> ::= <arraydecl> {; <arraydecl>}*
;;	   <arraydecl> ::= <var> [<exp>]

(define cs342:next-symbol
  (let ((c 0))
    (lambda ()
      (set! c (+ c 1))
      (string->symbol (string-append "g" (number->string c))))))

(define syntax-expand
  ; TYPE: (-> (parsed-exp) parsed-exp)
  (lambda (parsed-exp)
    (variant-case parsed-exp
      (lit () parsed-exp)
      (varref () parsed-exp)
      (app (rator rands)
        (let ((proc (syntax-expand rator))
	      (args (map syntax-expand rands)))
	  (make-app proc args)))
      (let (decls body)
	(let ((vars (map decl->var decls))
              (exps (map (lambda (decl)
			   (syntax-expand (decl->exp decl)))
			 decls)))
	  (make-app (make-proc vars (syntax-expand body))
		    exps)))
      (if (test-exp then-exp else-exp)
	  (make-if (syntax-expand test-exp)
		   (syntax-expand then-exp)
		   (syntax-expand else-exp)))
      (proc (formals body) 
	(make-proc formals (syntax-expand body)))
      (varassign (var exp)
	(make-varassign var (syntax-expand exp)))
      (begin (exp1 exp2)
	(make-begin (syntax-expand exp1)
		    (syntax-expand exp2)))
      (letmutable (decls body)
	 (make-letmutable (map cs342:syntax-expand-decl decls)
			  (syntax-expand body)))
      (letrec (decls body)
	(let ((vars (map decl->var decls))
              (exps (map decl->exp decls)))
	  (syntax-expand
	   (make-let
	    (map make-decl vars (map
				 (lambda (ignored) (make-lit 0))
				 exps))
	    (make-begin	       
	     (letrec ((new-vars (map (lambda (ignored)
				       (cs342:next-symbol))
				     vars))
		      (make-begins
		       ; TYPE: (-> ((list <var>) (list <exp>)) <exp>)
		       (lambda (vlist elist)
			 (if (null? vlist)
			     (make-lit 0)
			     (make-begin
			      (make-varassign (car vlist)
					      (car elist))
			      (make-begins (cdr vlist) (cdr elist)))))))
	       (make-let (map make-decl new-vars exps)
			 (make-begins vars
				      (map make-varref new-vars))))
	     body)))))
      (letrecproc (procdecls body)
	(let ((vars (map procdecl->var procdecls))
	      (formals-list (map procdecl->formals procdecls))
              (bodies (map procdecl->body procdecls)))
	  (syntax-expand
	   (make-letrec
	    (map make-decl vars (map make-proc formals-list bodies))
	    body))))
      (dynassign (var exp body)
	 (make-dynassign var (syntax-expand exp) (syntax-expand body)))
      (letdynamic (decls body)
	  (make-letdynamic (map cs342:syntax-expand-decl decls)
			   (syntax-expand body)))
      (letarray (arraydecls body)
	  (make-letarray (map cs342:syntax-expand-decl arraydecls)
			 (syntax-expand body)))
      (arrayref (array index)
	 (make-arrayref (syntax-expand array)
			(syntax-expand index)))
      (arrayassign (array index exp)
	 (make-arrayassign
	  (syntax-expand array)
	  (syntax-expand index)
	  (syntax-expand exp)))
      (letproc (procdecls body)
	 (let ((vars (map procdecl->var procdecls))
	       (formals-list (map procdecl->formals procdecls))
	       (bodies (map (lambda (procdecl)
			   (syntax-expand (procdecl->body procdecl)))
			 procdecls)))
	   (make-letproc (map make-procdecl vars formals-list bodies)
			 (syntax-expand body))))
      (local (decls body)
	  (make-local (map cs342:syntax-expand-decl decls)
		      (syntax-expand body)))
      (else
       (displayln "Warning: syntax-expand not expanding : " parsed-exp)
       parsed-exp))))

(define cs342:syntax-expand-decl
  ; TYPE: (-> (parsed-decl) parsed-decl)
  (lambda (decl)
    (make-decl (decl->var decl)
	       (syntax-expand (decl->exp decl)))))
