;;; $Id: type-check-polytype.scm,v 1.35 1999/06/23 16:57:22 leavens Exp $
;;; Type inference for Scheme programs
;;; AUTHORS: Curtis Clifton and Gary T. Leavens,
;;;          from an earlier verison by Steven L. Jenkins.

(load-quietly-from-lib "type-check-unify.scm")
(load-quietly-from-lib "type-check-syntax-helpers.scm")
(load-quietly-from-lib "type-check-global-var-types.scm")
(load-quietly-from-lib "type-check-type-translate.scm")
;;; (require 'pretty-print)

;;;--------------------------------------------------------------------------
;;;
;;; Error handling.

(define *tc:failure* ; TYPE: type-expr
  (tc:make-basic-type-expr 'type-error))

(define *tc:external-failure*
  (tc:type-translate *tc:failure*))

(define tc:error-type?
  ;; TYPE: (-> (type-expr) boolean)
  (lambda (typ)
    ((tc:occurs-in-type-expr? *tc:failure*) typ)))

(define tc:external-error-type?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (typ)
    (tc:occurs-in-external-type-expr? *tc:external-failure* typ)))

(define tc:external-list-error-type?
  ;; TYPE: (-> ((list external-type-expr)) boolean)
  (lambda (ls)
    (tc:some tc:external-error-type? ls)))

(define tc:fail 
  ;; TYPE: (-> (datum ...) poof)
  (lambda msg
    ;; EFFECT: print the error message
    (apply error msg)))

(define tc:nl-indent "\n  ")

(define tc:small-report
  ;; TYPE: (-> (string string external-type-expr external-type-expr) type-expr)
  (lambda (message expr left-desc left-type right-desc right-type)
    ;; EFFECT: print a small-error report, and return *tc:failure*
    (if (not (or (tc:external-error-type? left-type)
		 (tc:external-error-type? right-type)))
	;; only print the error if we haven't printed something for a subpart
	(tc:small-error
	 message
	 ": "
	 expr
	 (string-append tc:nl-indent left-desc ": ")
	 left-type
	 (string-append tc:nl-indent right-desc ": ")
	 right-type))
    *tc:failure*))

(define tc:application-small-report
  ;; TYPE: (-> (string string external-type-expr (list external-type-expr))
  ;;           symbol)
  (lambda (message expr left-desc left-type right-desc right-type-ls)
    ;; EFFECT: print a small-error report, and return *tc:failure*
    (if (not (or (tc:external-error-type? left-type)
		 (tc:external-list-error-type? right-type-ls)))
	;; only print the error if we haven't printed something for a subpart
	(tc:small-error
	 message
	 ": "
	 expr
	 (string-append tc:nl-indent left-desc ": ")
	 left-type
	 (string-append tc:nl-indent right-desc ": ")
	 right-type-ls))
    *tc:failure*))

(define tc:externalize
  ;; TYPE: (-> ((env uniq-sym type-expr) type-expr) external-type-expr)
  (lambda (e t)
    ;; (displayln "e: " e ", t: " t)
    (tc:type-translate (tc:subst-bindings e t))))

(define tc:resolve-bindings
  ;; TYPE: (-> ((env symbol binding) (env uniq-sym type-expr))
  ;;           (env symbol binding))
  (lambda (tenv unif-env)
    ;; EFFECT: return a type environment that is like tenv,
    ;;         but with all bound variables in the domain of unif-env changed
    ;;         to their bindings
    (tc:env-map (lambda (p)
		  (cons (car p)
			(tc:bind-type-expr
			 (tc:binding->binder (cdr p)))
			 ((tc:subst-bindings-c unif-env)
			  (tc:binding->type-expr (cdr p)))))
		tenv)))

(define tc:varref-type
  ;; TYPE: (-> (symbol (env symbol binding) (env uniq-sym type-expr))
  ;;           type-expr)
  (lambda (f typ-env e)
    ;; REQUIRES: f is a varref
    ;; ENSURES: result is a type for f, instantiated if f is generic
    ;;(displayln "tc:varref-type:")
    ;;(displayln "   f: " f)
    ;;(display   "   typ-env: ")
    ;;(pretty-print typ-env)
    ;;(display   "   e: ")
    ;;(pretty-print e)
    (if (tc:env-bound? f typ-env)
	(let ((binding (tc:env-value f typ-env)))
	  (let ((type (tc:binding->type-expr binding)))
	    (if (tc:bound? 'let binding)
		(tc:type-expr-instance (tc:subst-bindings e type))
		type)))
	(if (tc:env-bound? f *tc:global-var-types*)
	    (tc:type-expr-instance (tc:env-value f *tc:global-var-types*))
	    (begin
	      (tc:small-error (string-append
			       "unknown global name: "
			       (symbol->string f)))
	      *tc:failure*)))))

(define tc:quoted-constant-type 
  ;; TYPE: (-> (s-expression) type-expr)
  (lambda (x)
    ;; REQUIRES: x is a scheme expression with a constant type
    ;; ENSURES: result is an instance of the quoted constant type for x
    (if (eq? x '())
	(tc:make-applied-type-expr
	 (tc:make-basic-type-expr 'list) ; patch because #f = ()
	 (list (tc:new-variable-type-expr)))
	;; above was (tc:parse-type-expr '(list ?a))
	(tc:constant-type x))))

;;;(deftype tc:constant-type (-> (datum) type-expr)) ;??? experimental

(define tc:constant-type 
  ;; TYPE: (-> (s-expression) type-expr)
  (lambda (x)
    ;; REQUIRES: x is a scheme expression with a constant type
    ;; ENSURES: result is the type of the Scheme constant x
    (cond ((number? x)  (tc:make-basic-type-expr 'number))
	  ((char? x)    (tc:make-basic-type-expr 'char))
	  ((string? x)  (tc:make-basic-type-expr 'string))
	  ((boolean? x) (tc:make-basic-type-expr 'boolean))
	  ((symbol? x)  (tc:make-basic-type-expr 'symbol))
	  ((null? x)    (tc:make-applied-type-expr
			 (tc:make-basic-type-expr 'list)
			 (list (tc:new-variable-type-expr))))
	  ((list? x)
	   ;; the following only deals properly with flat lists
	   (let ((element-type (tc:constant-type (car x))))
	     (for-each (lambda (y)
			 (if (not (equal? (tc:constant-type y) element-type))
			     (set! element-type
				   (tc:make-basic-type-expr 'datum))))
		       (cdr x))
	     (tc:make-applied-type-expr
	      (tc:make-basic-type-expr 'list)
	      (list element-type))))
	  ((pair? x)
	   (tc:make-applied-type-expr
	    (tc:make-basic-type-expr 'pair)
	    (list (tc:constant-type (car x))
		  (tc:constant-type (cdr x)))))
	  ((vector? x)
	   (if (= (vector-length x) 0)
	       (tc:make-applied-type-expr
		(tc:make-basic-type-expr 'vector)
		(tc:new-variable-type-expr))
	       (let ((element-type (tc:constant-type (vector-ref x 0))))
		 (tc:vector-for-each
		  (lambda (y)
		    (if (not (equal? (tc:constant-type y) element-type))
			(set! element-type 'datum)))
		  x
		  1)
		 (tc:make-applied-type-expr
		  (tc:make-basic-type-expr 'vector)
		  (list element-type)))))
	  (else (tc:fail "unknown constant type for: " x)))))
  

;;; Type-checking of top-level defines requires passing an environment
;;; to tc:type.  This procedure maintains the old interface to avoid
;;; changing the existing interface for other calls.
(define tc:type
  ;; TYPE: (-> (s-expression) type-expr)
  (lambda (f)
    (set! tc:e-env (tc:env-empty))
    (tc:type-aux f (tc:env-empty))))

(define tc:type-aux
  ;; TYPE: (-> (s-expression (env symbol binding)) type-expr)
  (lambda (f typ-env)
    ;; ENSURES: result is the type of expression f
    (if (>= *tc:tracing-level* 1)
	(begin
	  (displayln "entering tc:type: ")
	  (pretty-print f)))
    (letrec
	((tc:match-given-type-general
	  ;; TYPE: (-> (type-expr string string string)
	  ;;           (-> (datum s-expression) type-expr))
	  (lambda (expected error-msg expected-label inferred-label)
	    (lambda (what xp)
	      ;; MODIFIES: tc:e-env
	      ;; EFFECT: if the type of xp is not expected,
	      ;;         then issue a small-error using what, error-msg,
	      ;;         expected-label and inferred-label
	      ;;         and return *tc:failure*
	      ;;         otherwise update tc:e-env and return the expected
	      (let* ((xp-type (j xp))
		     (unif-res
		      (tc:unify expected
				xp-type
				tc:e-env
				(tc:get-vars typ-env))))
		(if (not (tc:unification-error? unif-res))
		    (begin
		      (set! tc:e-env (something->value unif-res))
		      expected)
		    (begin
		      (tc:small-report
		       error-msg
		       what
		       expected-label
		       (tc:externalize tc:e-env expected)
		       inferred-label
		       (tc:externalize tc:e-env xp-type))
		      *tc:failure*))))))
	 (tc:match-given-type
	  ;; TYPE: (-> (type-expr string) (-> (s-expression) type-expr))
	  (lambda (expected error-msg)
	    (lambda (xp)
	      ((tc:match-given-type-general
		expected
		error-msg
		"Expected"
		"Inferred")
	       xp xp))))
	 (tc:check-for-void		; TYPE: (-> (s-expression) void)
	  (lambda (xp)
	    (tc:ignore
	     ((tc:match-given-type
	       (tc:make-basic-type-expr 'void)
	       "Expression in begin with ignored result has non-void type")
	      xp))))
	 (tc:begin-body-type
	  ;; TYPE: (-> ((list s-expression)) type-expr)
	  (lambda (body)
	    ;; REQUIRES: body is not empty
	    ;; MODIFIES: tc:e-env
	    ;; ENSURES: result is the type of the last s-expression in body,
	    ;;          and each of the others must have type void
	    ;; Note: type errors in all but the last expression do not cause
	    ;;       the type of the whole begin expression to be in error
	    (for-each tc:check-for-void (tc:but-last body))
	    (j (tc:last-item body))))
	 (tc:check-for-boolean	; TYPE: (-> (s-expression) void)
	  (lambda (xp)
	    (tc:ignore
	     ((tc:match-given-type
	       (tc:make-basic-type-expr 'boolean)
	       "Predicate in if expression has wrong type")
	      xp))))
	 (tc:must-match-type
	  ;; TYPE: (-> (string string string)
	  ;;           (-> (s-expression s-expression) type-expr))
	  (lambda (context-msg left-msg right-msg)
	    (lambda (left-exp right-exp)
	      ;; MODIFIES: tc:e-env
	      ;; EFFECT: if the types of left-exp and right-exp do not unify,
	      ;;         then issue an error using context-msg, left-msg,
	      ;;         and right-msg  and return *tc:failure*,
	      ;;         otherwise update tc:e-env and return the type
	      (let* ((left-type (j left-exp))
		     (right-type (j right-exp)))
		(if (or (tc:error-type? left-type)
			(tc:error-type? right-type))
		    *tc:failure*
		    (begin
		      (let ((unif-res (tc:unify left-type
						right-type
						tc:e-env
						(tc:get-vars typ-env))))
			(if (not (tc:unification-error? unif-res))
			    (begin
			      (set! tc:e-env (something->value unif-res))
			      left-type)
			    (let ((left-type-ext (tc:externalize
						  tc:e-env
						  left-type))
				  (right-type-ext (tc:externalize
						   tc:e-env
						   right-type)))
			      (if (not (or (tc:external-error-type?
					    left-type-ext)
					   (tc:external-error-type?
					    right-type-ext)))
				  (tc:small-error
				   context-msg
				   (string-append tc:nl-indent left-msg ": ")
				   left-exp
				   (string-append tc:nl-indent right-msg ": ")
				   right-exp
				   (string-append tc:nl-indent
						  left-msg
						  "'s type: ")
				   left-type-ext
				   (string-append tc:nl-indent
						  right-msg
						  "'s type: ")
				   right-type-ext))
			      *tc:failure*)))))))))
	 (tc:if-arms-must-match
	  ;; TYPE: (-> (s-expression s-expression) type-expr)
	  (lambda (left right)
	    ((tc:must-match-type
	      "Arms of if expression have different types"
	      "Left  arm"
	      "Right arm")
	     left right)))
	 (tc:check-if-statement	; TYPE: (-> (s-expression) type-expr)
	  (lambda (body)
	    ((tc:match-given-type
	      (tc:make-basic-type-expr 'void)
	      "Body of single-arm if statement has non-void type")
	     body)))
	 (tc:letrec-body-type
	  ;; TYPE: (-> (string type-expr s-expression) type-expr)
	  (lambda (name expected body)
	    ((tc:match-given-type-general
	      expected
	      "Mismatch between types of uses and defining expression for"
	      "Type of uses               "
	      "Type of defining expression")
	     name body)))
	 (j				; TYPE: (-> (s-expression) type-expr)
	  (lambda (f)
	    ;; MODIFIES: tc:e-env
	    ;; ENSURES: result is the type of f, once variables in tc:e-env
	    ;;          are substituted for the variables in result
	    ;; Note: This is a modified version of algorithm 'j'
	    ;;    from Robin Milner's paper, A Theory of Type Polymorphism in
	    ;;    Programming, JCSS, 17(3):348-375, Dec, 1978.
	    (if (>= *tc:tracing-level* 2)
		(begin
		  (displayln "entering j for " f)
		  (displayln "  typ-env is " typ-env)))
	    (cond
	     ((symbol? f)
	      (tc:varref-type f typ-env tc:e-env))
	     ((not (pair? f))
	      (tc:type-expr-instance (tc:constant-type f)))
	     ((not (list? f))
	      (tc:fail "bad syntax: " f))
	     ;; now we know that f is a non-empty list...
	     ((eq? (car f) 'deftype);; added for define-record desugaring
	      (if (not (= 3 (length f)))
		  (tc:fail "wrong number of arguments for deftype: " f)
		  (begin
		    ;; (displayln "deftype case " f)
		    (set! typ-env (tc:env-update
				   (cadr f)
				   (tc:bind-type-expr
				    'let 
				    (tc:type-untranslate (caddr f)))
				   typ-env))
		    (tc:make-basic-type-expr 'void))))
	     ((eq? (car f) 'quote)
	      (if (not (= 2 (length f)))
		  (tc:fail "wrong number of arguments for quote: " f)
		  (tc:quoted-constant-type (cadr f))))
	     ((eq? (car f) 'begin)
	      (let ((body (cdr f)))
		(if (null? body)
		    (tc:fail "no arguments to begin: " f)
		    (tc:begin-body-type (cdr f)))))
	     ((eq? (car f) 'if)
	      ;; (displayln "if is " f)
	      ;; (displayln "typ-env is " typ-env)
	      (let ((len (length f)))
		(if (or (< len 3) (> len 4))
		    (begin
		      (tc:small-error "Wrong number of arguments to if: " f)
		      *tc:failure*)
		    (begin
		      (tc:check-for-boolean (cadr f))
		      (if (= len 4)
			  (tc:if-arms-must-match (caddr f) (cadddr f))
			  (tc:check-if-statement (caddr f)))))))
	     ((eq? (car f) 'lambda)
	      (cond
	       ((< 3 (length f))
		(tc:fail "Bad syntax in lambda: " f))
	       ((symbol? (tc:lambda->args f))
		;; This case handles a single parameter, variable-arity lambda
		;; expression
		(let* ((parm (tc:lambda->args f))
		       (parm-type-var (tc:new-variable-type-expr))
		       (parm-type (tc:make-applied-type-expr
				   (tc:make-basic-type-expr 'list)
				   (list parm-type-var)))
		       (old-typ-env typ-env))
		  (set! typ-env
			(tc:env-update parm
				       (tc:bind-type-expr 'lambda parm-type)
				       typ-env))
		  (let ((body-type (j (tc:lambda->body f))))
		    (set! typ-env old-typ-env)
		    (if (not (tc:error-type? body-type))
			(tc:make-function-type-expr
			 (list parm-type-var (tc:make-basic-type-expr '...))
			 body-type)
			*tc:failure*))))
	       ((not (list? (tc:lambda->args f)))
		;; suppress type checking in this case (multiple parameters,
		;; variable-arity lambda expression)
		(tc:type-expr-instance (tc:make-function-type-expr
					(list (tc:make-basic-type-expr 'datum)
					      (tc:make-basic-type-expr '...))
					(tc:new-variable-type-expr))))
	       (else
		;; fixed-arity lambda expression
		(let* ((parms (tc:lambda->args f))
		       (parm-types
			(map (lambda (x) (tc:new-variable-type-expr)) parms))
		       (old-typ-env typ-env))
		  (set! typ-env (tc:env-update*
				 parms
				 (map (lambda (x)
					(tc:bind-type-expr 'lambda x))
				      parm-types)
				 typ-env))
		  (let ((body-type (j (tc:lambda->body f))))
		    (set! typ-env old-typ-env)
		    (if (not (tc:error-type? body-type))
			(tc:make-function-type-expr parm-types body-type)
			*tc:failure*))))))
	     ((eq? (car f) 'let)
	      (if (not (and (= 3 (length f))
			    (tc:list-of-2-elem-lists? (cadr f))))
		  (tc:fail "Bad syntax in let: " f)
		  (let ((decls (tc:let->bindings f))
			(old-typ-env typ-env))
		    (set! typ-env
			  (tc:env-update* (map car decls)
					  (map (lambda (x)
						 (tc:bind-type-expr
						  'let
						  (j (cadr x))))
					       decls)
					  typ-env))
		    (let ((res (j (tc:let->body f))))
		      (set! typ-env old-typ-env)
		      res))))
	     ;; A new case to handle defines
	     ((eq? (car f) 'define)
	      (if (not (and (= 3 (length f))
			    (symbol? (tc:second f))))
		  (tc:fail "Bad syntax in define: " f)
		  (let* ((name (tc:second f))
			 (body (tc:third f))
			 (expected (tc:new-variable-type-expr))
			 (expected-as-us
			  (tc:variable-type-expr->uniq-sym expected)))
		    (if (tc:env-bound? name typ-env)
			(set! tc:e-env
			      (tc:env-update
			       expected-as-us
			       (tc:binding->type-expr
				(tc:env-value name typ-env))
			       tc:e-env))
			(set! typ-env
			      (tc:env-update
			       name
			       (tc:bind-type-expr 'letrec expected)
			       typ-env)))
		    (let ((res (tc:letrec-body-type name expected body)))
		      (if (tc:error-type? res)
			  (set! tc:e-env
				(tc:env-update
				 expected-as-us
				 *tc:failure*
				 tc:e-env))
			  (set! tc:e-env
				(tc:env-update
				 expected-as-us
				 (tc:subst-bindings tc:e-env expected)
				 tc:e-env)))
		      res))))
	     ;; The case for letrec follows Case 6 in Cardelli's paper
	     ;; Basic Polymorphic Typechecking,
	     ;; Science of Computer Programming, 8(2):147-172, Apr 1987
	     ;; (i.e., type-inference for let's with recursive declarations)
	     ((eq? (car f) 'letrec)
	      (if (not (and (= 3 (length f))
			    (tc:list-of-2-elem-lists? (cadr f))))
		  (tc:fail "Bad syntax in letrec: " f)
		  (let* ((decls (tc:letrec->bindings f))
			 (typ-env-rec
			  (tc:env-update* (map car decls)
					  (map (lambda (x)
						 (tc:bind-type-expr
						  'letrec
						  (tc:new-variable-type-expr)))
					       decls)
					  (tc:env-empty)))
			 (old-typ-env typ-env))
		    (set! typ-env (tc:env-join typ-env-rec typ-env))
		    (for-each
		     (lambda (decl)
		       (let* ((name (car decl))
			      (body (cadr decl))
			      (expected
			       (tc:binding->type-expr (tc:env-value name
								    typ-env)))
			      (res (tc:letrec-body-type name expected body)))
			 (if (tc:error-type? res)
			     (set! tc:e-env
				   (tc:env-update
				    (tc:variable-type-expr->uniq-sym expected)
				    *tc:failure*
				    tc:e-env)))))
		     decls)
		    (let ((typ-env-rec2
			   ;; env with letrec vars let-bound to types
			   (tc:env-map
			    (lambda (x)
			      (let* ((var (car x))
				     (binding (cdr x))
				     (newexp
				      (tc:subst-bindings
				       tc:e-env
				       (tc:binding->type-expr binding))))
				(cons var (tc:bind-type-expr 'let newexp))))
			    typ-env-rec)))
		      (set! typ-env (tc:env-join typ-env-rec2
						 old-typ-env))
		      (let ((res (j (tc:letrec->body f))))
			(set! typ-env old-typ-env)
			res)))))
	     ;; The next case is a hack to make direct application of lambdas
	     ;; have a polymorphic type; it's not really needed.
	     ((and (pair? (tc:appl->rator f))
		   (eq? (car (tc:appl->rator f)) 'lambda)
		   (pair? (cdr f)))
	      (j (list 'let (tc:map2 list
				     (tc:lambda->args (tc:appl->rator f))
				     (tc:appl->rands f))
		       (tc:lambda->body (tc:appl->rator f)))))
	     (else ;; applications, the general case
	      (let* ((fun (tc:appl->rator f))
		     (result-type (tc:new-variable-type-expr))
		     (oper-type (j fun))
		     (arg-types (map j (tc:appl->rands f))))
		(if (>= *tc:tracing-level* 2)
		    (begin
		      (display "For application ")
		      (pretty-print f)
		      (displayln "  result-type is " result-type
				 "  and fun is " fun)
		      (displayln "  oper-type is " oper-type)
		      (displayln "  and arg-types are " arg-types)))
		(let ((unification-result
		       (tc:unify
			oper-type
			(tc:make-function-type-expr arg-types result-type)
			tc:e-env
			(tc:get-vars typ-env))))
		  ;; (displayln "j: unification-result: " unification-result)
		  (if (tc:unification-error? unification-result)
		      (begin
			(tc:application-small-report
			 "Operator type and argument types do not match in"
			 f
			 "Operator type "
			 (tc:externalize tc:e-env oper-type)
			 "Argument type list"
			 (map (lambda (t)
				(tc:externalize tc:e-env t))
			      arg-types))
			*tc:failure*)
		      (begin
			(set! tc:e-env (something->value unification-result))
			result-type)))))
	     ))))			; end defn. of j and letrec bindings
      (let ((t (j f)))
	(if (>= *tc:tracing-level* 1)
	    (begin
	      (displayln "tc:type finished")
	      (displayln "  t is " t)
	      (displayln "  typ-env is ")
	      (pretty-print typ-env)
	      (displayln "  tc:e-env is ")
	      (pretty-print tc:e-env)))
	(let ((expanded-type (tc:subst-bindings tc:e-env t)))
	  (if (>= *tc:tracing-level* 1)
	      (displayln "  for " f " expanded-type is " expanded-type))
	  (if (tc:error-type? expanded-type)
	      *tc:failure*
	      expanded-type))))))

(define tc:remove-if-not-pred
  ;; TYPE: (-> ((-> (datum) boolean) (list T)) (list T))
  (lambda (pred ls)
    (cond
     ((null? ls) '())
     ((pred (car ls)) (cons (car ls) (tc:remove-if-not-pred pred (cdr ls))))
     (else (tc:remove-if-not-pred pred (cdr ls))))))

(define tc:get-vars
  ;; TYPE: (-> ((env symbol binding)) (list variable-type-expr))
  (lambda (env)
    ;; EFFECT: return a list of all the variable-type-expr bound in env
    (if (tc:env-empty? env)
	'()
	(let ((te (tc:binding->type-expr (cdr (tc:env-first env)))))
	  (if (tc:variable-type-expr? te)
	      (cons te (tc:get-vars (tc:env-rest env)))
	      (tc:get-vars (tc:env-rest env)))))))

(define tc:list-of-2-elem-lists?
  ;; TYPE: (-> (datum) boolean)
  (lambda (x)
    (and (list? x)
	 (or (null? x)
	     (and (list? (car x))
		  (= 2 (length (car x)))
		  (tc:list-of-2-elem-lists? (cdr x)))))))
