;;; $Id: type-check-type-expr.scm,v 1.4 1999/06/23 16:57:22 leavens Exp $
;;;
;;; Internal types in the type checker:
;;;
;;;   <type-expr> ::= <variable-type-expr>
;;;            | <basic-type-expr>
;;;            | <function-type-expr>
;;;            | <and-type-expr>
;;;            | <applied-type-expr>

(define-record tc:variable-type-expr (name level))
(define-record tc:basic-type-expr (symbol))
(define-record tc:function-type-expr (arg-types result-type))
(define-record tc:and-type-expr (conjoined-types))
(define-record tc:applied-type-expr (operator-type operand-types))

(load-quietly-from-lib "type-check-utils.scm")

(define tc:type-expr?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (variant-case d
      (tc:variable-type-expr () #t)
      (tc:basic-type-expr () #t)
      (tc:function-type-expr (arg-types result-type)
			     (and (tc:type-expr? result-type)
				  (tc:all tc:type-expr? arg-types)))
      (tc:and-type-expr (conjoined-types)
			(tc:all tc:type-expr? conjoined-types))
      (tc:applied-type-expr (operator-type operand-types)
			     (and (tc:basic-type-expr? operator-type)
				  (tc:all tc:type-expr? operand-types)))
      (else #f))))

(define tc:non-variable-basic-type-expr?
  ;; TYPE: (-> (type-expr) boolean)
  ;; REQUIRES: d is not a type variable
  ;; ENSURES: result is true just when d is a basic type
  tc:basic-type-expr?)

(define tc:poof-type-expr?
  ;; TYPE: (-> (type-expr) boolean)
  (lambda (d)
    (variant-case d
      (tc:basic-type-expr (symbol) (eq? symbol 'poof))
      (else #f))))

(define tc:datum-type-expr?
  ;; TYPE: (-> (type-expr) boolean)
  (lambda (d)
    (variant-case d
      (tc:basic-type-expr (symbol) (eq? symbol 'datum))
      (else #f))))

(define tc:dots-type-expr?
  ;; TYPE: (-> (type-expr) boolean)
  (lambda (d)
    (variant-case d
      (tc:basic-type-expr (symbol) (eq? symbol '...))
      (else #f))))

(define tc:varargs-function-type-expr?
  ;; TYPE: (-> (type-expr) boolean)
  (lambda (d)
    (variant-case d
      (tc:function-type-expr (arg-types)
	 (and (not (null? arg-types))
	      (tc:dots-type-expr? (tc:last-item arg-types))))
      (else #f))))

(define tc:make-variable-type-expr
  ;; TYPE: (-> (symbol) type-expr)
  (lambda (s) (make-tc:variable-type-expr s 0)))

(define tc:variable-type-expr-new-level
  ;; TYPE: (-> (type-expr number) type-expr)
  (lambda (t n)
    ;; REQUIRES: t is a variable-type-expr
    ;; EFFECT: returns a copy of t with the level field set to n
    (variant-case t
      (tc:variable-type-expr (name) (make-tc:variable-type-expr name n))
      (else
       (error "tc:variable-type-expr-new-level: not a variable-type-expr: " t)
       ))))

(define tc:new-variable-type-expr
  ;; TYPE: (-> () type-expr)
  (let ((variable-count 0))
    (lambda ()
      ;; EFFECT: returns a unique variable-type-expr with the name
      ;;         ?variable-count
      (set! variable-count (+ variable-count 1))
      (tc:make-variable-type-expr 
       (string->symbol (string-append "?" (number->string variable-count)))))))

(define tc:make-basic-type-expr
  ;; TYPE: (-> (symbol) type-expr)
  make-tc:basic-type-expr)

(define tc:make-function-type-expr
  ;; TYPE: (-> ((list type-expr) type-expr) type-expr)
  make-tc:function-type-expr)
  
(define tc:make-and-type-expr
  ;; TYPE: (-> ((list type-expr)) type-expr)
  make-tc:and-type-expr)

(define tc:make-applied-type-expr
  ;; TYPE: (-> (type-expr (list type-expr)) type-expr)
  make-tc:applied-type-expr)

(define tc:variable-type-expr->uniq-sym
  ;; TYPE: (-> (type-expr) uniq-sym)
  (lambda (vt)
    ;; REQUIRES: vt is a variable-type-expr
    ;; EFFECT: returns a symbol formed by concatenating vt's name, \#, and
    ;;         vt's level
    (variant-case vt
      (tc:variable-type-expr (name level)
			     (string->symbol
			      (string-append (symbol->string name)
					     "#"
					     (number->string level))))
      (else
       (error "tc:variable-type-expr->uniq-sym: not a variable-type-expr: " vt)
       ))))

(define tc:uniq-sym->variable-type-expr
  ;; TYPE: (-> (uniq-sym) type-expr)
  (letrec ((findit
	    ;; TYPE: (-> (string number) number)
	    (lambda (s p)
	      ;;(displayln "entering findit: s " s " p " p)
	      (cond
	       ((= p (string-length s)) -1)
	       ((equal? (string-ref s p) #\#) p)
	       (else (findit s (+ p 1)))))))
    (lambda (us)
      ;; EFFECT: returns a variable-type-expr with the appropriate name and
      ;;         level extracted from us
      (let* ((ustr (symbol->string us))
	     (pos (findit ustr 1)))	; don't check position 0, it can't
					; (legally) be there
	(if (not (= pos -1))
	    (make-tc:variable-type-expr
	     (string->symbol (substring ustr 0 pos))
	     (string->number (substring ustr (+ pos 1) (string-length ustr))))
	    (error "tc:uniq-sym->variable-type-expr: invalid uniq-sym: "
		   us))))))

(define tc:parse-type-expr
  (letrec ((symbol-is-variable?
	    ;; TYPE: (-> (symbol) boolean)
	    (lambda (x)
	      (char=? (string-ref (symbol->string x) 0) #\?))))
    ;; TYPE: (-> (datum) type-expr)
    (lambda (d)
      ;; REQUIRES: d follows the syntax of a valid external type-expr
      ;; EFFECT: returns the type-expr described by d
      (cond
       ;; a variable-type-expr is a symbol that begins with "?"
       ((and (symbol? d) (symbol-is-variable? d))
	(tc:make-variable-type-expr d))
       ;; a basic-type-expr is a symbol that is not a variable
       ((symbol? d) (tc:make-basic-type-expr d))
       ;; a function-type-expr, an and-type-expr, and an applied-type-expr
       ;; all are lists of length 2 or more
       ((and (list? d) (>= (length d) 2))
	(cond
	 ;; a function-type-expr is of the form: (-> ({<type-expr>}*) <type-expr>)
	 ((eq? (car d) '->)		
	  (if (and (list? (cadr d)) (= (length d) 3))
	      (tc:make-function-type-expr
	       (map tc:parse-type-expr (cadr d))
	       (tc:parse-type-expr (caddr d)))
	      (error "tc:parse-type-expr: invalid function type syntax: " d)))
	 ;; an and-type-expr is of the form: (and {<type-expr>}+)
	 ((eq? (car d) 'and)
	  (tc:make-and-type-expr (map tc:parse-type-expr (cdr d))))
	 ;; an applied-type-expr is of the form:
	 ;; (<basic-type-expr> {<type-expr>}+)
	 ((and (symbol? (car d)) (not (symbol-is-variable? (car d))))
	  (tc:make-applied-type-expr (tc:parse-type-expr (car d))
				     (map tc:parse-type-expr (cdr d))))
	 (else (error "tc:parse-type-expr: syntax error in " d))))
       (else (error "tc:parse-type-expr: syntax error in " d))))))
    
(define tc:occurs-in-type-expr?
  ;; TYPE: (-> (type-expr) (-> (type-expr) boolean))
  (lambda (typ1)
    (lambda (typ2)
      ;; EFFECT: determines if typ1 occurs anywhere in typ2
      (or
       (equal? typ1 typ2)
       (variant-case typ2
	 (tc:variable-type-expr () #f)
	 (tc:basic-type-expr () #f)
	 (tc:function-type-expr (arg-types result-type)
				(or ((tc:occurs-in-type-expr? typ1)
				     result-type)
				    (tc:some (tc:occurs-in-type-expr? typ1)
					     arg-types)))
	 (tc:and-type-expr (conjoined-types)
			   (tc:some (tc:occurs-in-type-expr? typ1)
				    conjoined-types))
	 (tc:applied-type-expr (operator-type operand-types)
			       (or ((tc:occurs-in-type-expr? typ1)
				    operator-type)
				   (tc:some (tc:occurs-in-type-expr? typ1)
					    operand-types)))
	 (else
	  (error "tc:occurs-in-type-expr?: not a valid type-expr: " typ2)))))))

(define tc:type-expr-uniquify
  ;; TYPE: (and (-> (type-expr) type-expr)
  ;;            (-> ((pair type-expr type-expr)) (pair type-expr type-expr))))
  (let ((*tc:unique-var-number* 0))
    (letrec ((uniquify
	      ;; TYPE: (-> (type-expr) type-expr)
	      (lambda (te)
		(variant-case te
		  (tc:variable-type-expr ()
					 (tc:variable-type-expr-new-level
					  te *tc:unique-var-number*))
		  (tc:basic-type-expr () te)
		  (tc:function-type-expr (arg-types result-type)
					 (tc:make-function-type-expr
					  (map uniquify arg-types)
					  (uniquify result-type)))
		  (tc:and-type-expr (conjoined-types)
				    (tc:make-and-type-expr
				     (map uniquify conjoined-types)))
		  (tc:applied-type-expr (operator-type operand-types)
					(tc:make-applied-type-expr
					 (uniquify operator-type)
					 (map uniquify operand-types)))
		  (else
		   (error "tc:type-expr-uniquify: not a valid type-expr: "
			  te))))))
      (lambda (type-expr)
	;; ENSURES: result is a type-expr, or pair of type-expr, that is
	;;          like type-expr, but in which the type variables are
	;;          unique (in this execution)
	;; NOTE: This procedure must handle pairs of type-expr so that
	;;       the binding of variables between the abstract and
	;;       representation types in a defrep can be maintained.
	(set! *tc:unique-var-number* (+ *tc:unique-var-number* 1))
	(if (pair? type-expr)
	    (cons (uniquify (car type-expr))
		  (uniquify (cdr type-expr)))
	    (uniquify type-expr))))))

(define tc:type-expr-instance
  ;; TYPE: (-> (type-expr) type-expr)
  (lambda (te)
    ;; ENSURES: result is a type-expr that is
    ;;          like te, but in which the type variables are
    ;;          instantiated using unique variable-type-expr
    (let ((env
	   ;; TYPE: (env uniq-sym type-expr)
	   (tc:env-empty)))
      (letrec
	  ((instancer
	    ;; TYPE: (-> (type-expr) type-expr)
	    (lambda (te)
	      (variant-case te
		(tc:variable-type-expr ()
		  (let ((us (tc:variable-type-expr->uniq-sym te)))
		    (if (tc:env-bound? us env)
			(tc:env-value us env)
			(let ((var (tc:new-variable-type-expr)))
			  (set! env (tc:env-update us var env))
			  var))))
		(tc:basic-type-expr () te)
		(tc:function-type-expr (arg-types result-type)
				       (tc:make-function-type-expr
					(map instancer arg-types)
					(instancer result-type)))
		(tc:and-type-expr (conjoined-types)
				  (tc:make-and-type-expr
				   (map instancer conjoined-types)))
		(tc:applied-type-expr (operator-type operand-types)
				      (tc:make-applied-type-expr
				       (instancer operator-type)
				       (map instancer operand-types)))
		(else
		 (error "tc:type-expr-instance: not a valid type-expr: "
			te))))))
	;; (displayln "tc:type-expr-instance: te: ")
	;; (pretty-print te)
	(instancer te)))))
  
(define tc:variable-type-expr-eq?
  ;; TYPE: (-> (type-expr type-expr) boolean)
  (lambda (v1 v2)
    ;; REQUIRES: v1 and v2 are variable-type-expr
    ;; EFFECT: returns true if v1 match on their name and level fields
    (and (eq? (tc:variable-type-expr->name v1)
	      (tc:variable-type-expr->name v2))
	 (= (tc:variable-type-expr->level v1)
	    (tc:variable-type-expr->level v2)))))

(define tc:basic-type-expr-eq?
  ;; TYPE: (-> (type-expr type-expr) boolean)
  (lambda (typ1 typ2)
    ;; REQUIRE: typ1 and typ2 are basic-type-exprs
    ;; EFFECT: returns true if typ1 and typ2 are equivalent
    (eq? (tc:basic-type-expr->symbol typ1)
	 (tc:basic-type-expr->symbol typ2))))

(define tc:binding?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (and (pair? d)
	 (or (eq? 'let (car d))
	     (eq? 'letrec (car d))
	     (eq? 'lambda (car d)))
	 (tc:type-expr? (cdr d)))))

(define tc:bind-type-expr
  ;; TYPE: (-> (symbol type-expr) binding)
  (lambda (s te)
    (cons s te)))

(define tc:bound?
  ;; TYPE: (-> (symbol binding) boolean)
  (lambda (s b)
    (eq? (car b) s)))

(define tc:binding->type-expr
  ;; TYPE: (-> (binding) type-expr)
  cdr)

(define tc:binding->binder
  ;; TYPE: (-> (binding) symbol)
  car)
